#!/usr/bin/perl -w
#
# MD5 implemented in AMD CAL IL.
#
# Author: Marc Bevand <bevand_m (at) epita.fr>
#

use strict;

# pick a strategy for implementing shifting
my $shift_strategy = "c";
my %map_shft_literals = ();
my $code;
my $enable_bfi_int;
my $elm_per_threads = 12; # keep in sync with ELM_PER_THREAD in C code
my ($d1, $d2, $d3) = qw/r3 r27 r34/;
my ($targetd1, $targetd2, $targetd3) = qw/r23 r30 r37/;
my ($firstword1, $firstword2, $firstword3) = qw/r7 r28 r35/;
my ($secondword1, $secondword2, $secondword3) = qw/r8 r29 r36/;
my $zero = 'l103.yyyy';
my $one = 'l100.wwww';

# Returns word X[k] processed during the specified step for the given
# interleaved MD5.
sub step_to_x_k
{
    my ($step, $interleave) = @_;
    my @elm = qw/xxxx yyyy zzzz wwww/;
    my $k;
    die "*bug*: step cannot be $step\n" if $step >= 64;
    if ($step < 16)      { $k = $step; }
    elsif ($step < 16*2) { $k = (5 * $step + 1) % 16; }
    elsif ($step < 16*3) { $k = (3 * $step + 5) % 16; }
    else                 { $k = (7 * $step) % 16; }
    if ($k <= 1) {
	if ($interleave == 0) {
	    return (!$k) ? $firstword1 : $secondword1;
	} elsif ($interleave == 1) {
	    return (!$k) ? $firstword2 : $secondword2;
	} elsif ($interleave == 2) {
	    return (!$k) ? $firstword3 : $secondword3;
	} else {
	    die "*bug*: unknown interleave $interleave";
	}
    }
    return sprintf "r%i", 9 + ($k - 2);
}

# Returns the sine value T_i used in the specified step (cb1[?].?).
sub step_to_t_i
{
    my ($step) = @_;
    my @elm = qw/xxxx yyyy zzzz wwww/;
    die "*bug*: step cannot be $step\n" if $step >= 64;
    return sprintf "cb1[%i].%s", $step / 4, $elm[$step % 4];
}

# Returns the register corresponding to A,B,C,D for the given interleaved MD5.
sub abcd_to_r0123
{
    my $interleave = $_[4];
    my %h1 = (A => "r0", B => "r1", C => "r2", D => "r3");
    my %h2 = (A => "r24", B => "r25", C => "r26", D => "r27");
    my %h3 = (A => "r31", B => "r32", C => "r33", D => "r34");
    for (my $i = 0; $i < 4; $i++) {
	my $h = (!$interleave ? \%h1 : (($interleave == 1) ? \%h2 : \%h3));
	die "*bug*: invalid variable: $_[$i]\n" if !exists($h->{$_[$i]});
	$_[$i] = $h->{$_[$i]};
    }
}

# Declares constant used by shifting operations.
sub declare_shft_literals
{
    my @literals = ();
    my @elm = qw/xxxx yyyy zzzz wwww/;
    for my $s (
	7, 12, 17, 22, # round 1
	5, 9, 14, 20,  # round 2
	4, 11, 16, 23, # round 3
	6, 10, 15, 21, # round 4
    ) {
	if ($shift_strategy eq 'a') {
	    push @literals, 32 - $s, $s;
	} elsif ($shift_strategy eq 'b') {
	    push @literals, 32 - $s, 2 ** $s;
	} elsif ($shift_strategy eq 'c') {
	    push @literals, 32 - $s, $s;
	} else {
	    die "*bug*: invalid shift strategy: $shift_strategy\n";
	}
    }
    $code .= sprintf "  ; shift literals\n";
    for (my $i = 0; $i < scalar(@literals); $i++) {
	my $v = $literals[$i];
	my $lx = sprintf "l%d", $i / 4;
	my $e = $elm[$i % 4];
	$map_shft_literals{$v} = "$lx.$e";
	if (!($i % 4)) {
	    $code .= sprintf "  dcl_literal %s, %d", $lx, $v;
	} elsif (($i % 4) != 3) {
	    $code .= sprintf ", %d", $v;
	} else {
	    $code .= sprintf ", %d\n", $v;
	}
    }
}

# Returns the literal having the specified value.
sub get_shft_literal
{
    my ($v) = @_;
    exists($map_shft_literals{$v}) or die "cannot find literal $v\n";
    return $map_shft_literals{$v};
}

# Rotates r5 with a given shift strategy, stores the output in r5. Multiple
# strategies are available:
# a) uses 3 instructions (RIGHT SHIFT by "32-s", LEFT SHIFT by "s",
#   OR the 2 results)
# b) uses 2 instructions (RIGHT SHIFT by "32-s", MAD with "2**s" and
#   "right shift result")
# c) uses 1 instruction (BITALIGN with "32-s")
sub execute_shift_strategy
{
    my ($s_lit, $s_lit_compl, $s_lit_power) = @_;
    if ($shift_strategy eq 'a') {
	$code .= <<EOF;
    ishl r6, r5, $s_lit
    ushr r5, r5, $s_lit_compl
    ior  r5, r6, r5
EOF
    } elsif ($shift_strategy eq 'b') {
	$code .= <<EOF;
    ushr r6, r5, $s_lit_compl
    umad r5, r5, $s_lit_power, r6
EOF
    } elsif ($shift_strategy eq 'c') {
	$code .= <<EOF;
    bitalign r5, r5, r5, $s_lit_compl
EOF
    } else {
	die "unknown shift strategy: $shift_strategy\n";
    }
}

# Implements a round 1 step: t = x + ((t + F(x,y,z) + X[k] + T_i) <<< s)
sub round1_step
{
    my ($step, $t, $x, $y, $z, $s, $interleave) = @_;
    my $x_k = step_to_x_k($step, $interleave);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z, $interleave);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 1)\n";
    # F(x,y,z) = (x AND y) OR (NOT(x) AND z)
    # Equivalent: ((y XOR z) AND x) XOR z
    if ($step == 0 or !$enable_bfi_int or $interleave == 2) {
	$code .=
	"    ixor r5, $y, $z\n".
	"    and  r5, r5, $x\n".
	"    ixor r5, r5, $z\n";
    } else {
	$code .=
	"    ibit_extract r5, $z, $y, $x\n";
    }
    $code .=
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + F() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

# Implements a round 2 step: t = x + ((t + G(x,y,z) + X[k] + T_i) <<< s)
sub round2_step
{
    my ($step, $t, $x, $y, $z, $s, $interleave) = @_;
    my $x_k = step_to_x_k($step, $interleave);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z, $interleave);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 2)\n";
    # G(x,y,z) = (x AND z) OR (y AND NOT(z))
    # Equivalent: ((x XOR y) AND z) XOR y
    if (!$enable_bfi_int or $interleave == 1) {
	$code .=
	"    ixor r5, $x, $y\n".
	"    and  r5, r5, $z\n".
	"    ixor r5, r5, $y\n";
    } else {
	$code .=
	"    ibit_extract r5, $y, $x, $z\n";
    }
    $code .=
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + G() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

# Implements a round 3 step: t = x + ((t + H(x,y,z) + X[k] + T_i) <<< s)
sub round3_step
{
    my ($step, $t, $x, $y, $z, $s, $interleave) = @_;
    my $x_k = step_to_x_k($step, $interleave);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z, $interleave);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    $code .= "\n".
    "    ; step $step (round 3)\n".
    # H(x,y,z) = x XOR y XOR z
    "    ixor r5, $y, $z\n".
    "    ixor r5, r5, $x\n".
    # X[k] + T_i
    "    iadd r6, $x_k, $t_i\n".
    # sum = t + H() + X[k] + T_i
    "    iadd r5, $t, r5\n".
    "    iadd r5, r5, r6\n";
    # sum <<< s
    execute_shift_strategy($s_lit, $s_lit_compl, $s_lit_power);
    # add x, store in t
    $code .= "    iadd $t, $x, r5\n";
}

# No need to implement a round 4 step:
# t = x + ((t + I(x,y,z) + X[k] + T_i) <<< s)
# Instead the reverse of a round 4 step is needed for reversing:
# t = ((t - x) >>> s) - I(x,y,z) - X[k] - T_i
sub rev_round4_step
{
    my ($step, $t, $x, $y, $z, $s, $interleave) = @_;
    my $x_k = step_to_x_k($step, $interleave);
    my $t_i = step_to_t_i($step);
    abcd_to_r0123($t, $x, $y, $z, $interleave);
    my $s_lit = get_shft_literal($s) if ($shift_strategy eq 'a' or
	$shift_strategy eq 'c');
    my $s_lit_compl = get_shft_literal(32 - $s);
    my $s_lit_power = get_shft_literal(2 ** $s) if ($shift_strategy eq 'b');
    # TODO: call execute_shift_strategy() instead of hardcoding bitalign
    $code .= "\n".
    "    ; reverse step $step (round 4)\n".
    # t - x
    "    inegate r5, $x\n".
    "    iadd r5, $t, r5\n".
    # rotate to the right
    "    bitalign r5, r5, r5, $s_lit\n".
    # I(x,y,z) = y XOR (x OR NOT(z))
    "    inot r6, $z\n".
    "    ior r6, $x, r6\n".
    "    ixor r6, $y, r6\n".
    # subsract the value of I(...)
    "    inegate r6, r6\n".
    "    iadd r5, r5, r6\n".
    # substract X[k]
    "    inegate r6, $x_k\n".
    "    iadd r5, r5, r6\n".
    # substract T_i, storing the final value in t
    "    inegate r6, $t_i\n".
    "    iadd $t, r5, r6\n";
}

sub generate_kernel
{
    my ($fname, $pos, $bfi) = @_;
    $enable_bfi_int = $bfi;
    $code = "";
    $code .= <<EOF;
il_cs
  dcl_num_thread_per_group %d
  ; message blocks
  dcl_cb cb0[4]
  ; sine values
  dcl_cb cb1[16]
EOF
    declare_shft_literals();
    $code .= <<EOF;
  ; l100.x number of iterations
  ; l100.y used to access g[], must be sizeof (thread_state_t) / 16
  ; l100.z -1, used in various places
  ; l100.w 1, used in various places
  dcl_literal l100, %d, %lu, -1, 1
  ; l101.xyzw initial intermediate hash values A, B, C, D
  dcl_literal l101, 0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476
  ; l102.xyzw target hash A, B, C, D
  dcl_literal l102, 0x%x, 0x%x, 0x%x, 0x%x
  ; l103.x mask to extract position in pos_wuid
  ; l103.y 0, used in various places
  ; l103.z 2, used in various places
  ; l103.w value of "s_nomore"
  dcl_literal l103, 0x%x, 0, 2, %d
  ; l104-l107 to implement charset
  ; l109.x mask to extract high wuid bits in first word of pos_wuid
  dcl_literal l104,       0xff, 0x%x, 0x%x, 0
  dcl_literal l105,     0xff00, 0x%x, 0x%x, 0
  dcl_literal l106,   0xff0000, 0x%x, 0x%x, 0
  dcl_literal l107, 0xff000000, 0x%x, 0, 0
  dcl_literal l109, 0x%x, 0, 0, 0

  ; r0-r3   A,B,C,D for first interleaved MD5
  ; r4.x    counter
  ; r4.y    offset to this thread's state in g[]
  ; r5-r6   temporary values
  ; r7-r8   message words X[0]-X[1] for first interleaved MD5
  ; r9-r22  message words X[2]-X[15]
  ; r23     target value D at step 45 for first interleaved MD5
  ; r24-r27 A,B,C,D for second interleaved MD5
  ; r28-r29 message words X[0]-X[1] for second interleaved MD5
  ; r30     target value D at step 45 for second interleaved MD5
  ; r31-r34 A,B,C,D for third interleaved MD5
  ; r35-r36 message words X[0]-X[1] for third interleaved MD5
  ; r37     target value D at step 45 for third interleaved MD5

  umul r4.y, vAbsTidFlat.x, l100.y

  ; load X[0]-X[1] for first interleaved MD5
  mov  $firstword1.x, g[r4.y+0].y
  mov  $secondword1.x, g[r4.y+0].z
  mov  $firstword1.y, g[r4.y+1].x
  mov  $secondword1.y, g[r4.y+1].y
  mov  $firstword1.z, g[r4.y+1].w
  mov  $secondword1.z, g[r4.y+2].x
  mov  $firstword1.w, g[r4.y+2].z
  mov  $secondword1.w, g[r4.y+2].w
  ; load X[0]-X[1] for second interleaved MD5
  mov  $firstword2.x, g[r4.y+3].y
  mov  $secondword2.x, g[r4.y+3].z
  mov  $firstword2.y, g[r4.y+4].x
  mov  $secondword2.y, g[r4.y+4].y
  mov  $firstword2.z, g[r4.y+4].w
  mov  $secondword2.z, g[r4.y+5].x
  mov  $firstword2.w, g[r4.y+5].z
  mov  $secondword2.w, g[r4.y+5].w
  ; load X[0]-X[1] for third interleaved MD5
  mov  $firstword3.x, g[r4.y+6].y
  mov  $secondword3.x, g[r4.y+6].z
  mov  $firstword3.y, g[r4.y+7].x
  mov  $secondword3.y, g[r4.y+7].y
  mov  $firstword3.z, g[r4.y+7].w
  mov  $secondword3.z, g[r4.y+8].x
  mov  $firstword3.w, g[r4.y+8].z
  mov  $secondword3.w, g[r4.y+8].w
  ; init X[2]-X[15]
  mov  r9, cb0[0].zzzz
  mov r10, cb0[0].wwww
  mov r11, cb0[1].xxxx
  mov r12, cb0[1].yyyy
  mov r13, cb0[1].zzzz
  mov r14, cb0[1].wwww
  mov r15, cb0[2].xxxx
  mov r16, cb0[2].yyyy
  mov r17, cb0[2].zzzz
  mov r18, cb0[2].wwww
  mov r19, cb0[3].xxxx
  mov r20, cb0[3].yyyy
  mov r21, cb0[3].zzzz
  mov r22, cb0[3].wwww

  ; Reverse MD5 steps to compute D after step 45.
  ;
  ; Given a message block of 16 32-bit words (x) where all the words are known
  ; except x[0], and a 128-bit hash value (hash), executes the last 15 steps of
  ; round 4 of MD5 in reverse (from step 63 down to and including step 49) to
  ; compute the known intermediate hash values from the oldest step possible.
  ; Out of these values, D comes from step 45 and is output into d_res. This
  ; function enables an MD5 brute forcer to be able to determine after executing
  ; only 46 steps (0-45) if the D value computed so far indicates a potential 
  ; candidate message match.

  ; Load A,B,C,D for first interleaved MD5
  mov r0, l102.xxxx
  mov r1, l102.yyyy
  mov r2, l102.zzzz
  mov r3, l102.wwww
  inegate r5, l101.xxxx
  iadd r0, r0, r5
  inegate r5, l101.yyyy
  iadd r1, r1, r5
  inegate r5, l101.zzzz
  iadd r2, r2, r5
  inegate r5, l101.wwww
  iadd r3, r3, r5
  ; Load A,B,C,D for second interleaved MD5
  mov r24, r0
  mov r25, r1
  mov r26, r2
  mov r27, r3
  ; Load A,B,C,D for third interleaved MD5
  mov r31, r0
  mov r32, r1
  mov r33, r2
  mov r34, r3
EOF
    my $s = 63;
    rev_round4_step($s  ,'B','C','D','A','21', 0); # step 63
    rev_round4_step($s  ,'B','C','D','A','21', 1);
    rev_round4_step($s--,'B','C','D','A','21', 2);
    rev_round4_step($s,  'C','D','A','B','15', 0);
    rev_round4_step($s,  'C','D','A','B','15', 1);
    rev_round4_step($s--,'C','D','A','B','15', 2);
    rev_round4_step($s,  'D','A','B','C','10', 0);
    rev_round4_step($s,  'D','A','B','C','10', 1);
    rev_round4_step($s--,'D','A','B','C','10', 2);
    rev_round4_step($s,  'A','B','C','D', '6', 0);
    rev_round4_step($s,  'A','B','C','D', '6', 1);
    rev_round4_step($s--,'A','B','C','D', '6', 2);
    rev_round4_step($s,  'B','C','D','A','21', 0);
    rev_round4_step($s,  'B','C','D','A','21', 1);
    rev_round4_step($s--,'B','C','D','A','21', 2);
    rev_round4_step($s,  'C','D','A','B','15', 0);
    rev_round4_step($s,  'C','D','A','B','15', 1);
    rev_round4_step($s--,'C','D','A','B','15', 2);
    rev_round4_step($s,  'D','A','B','C','10', 0);
    rev_round4_step($s,  'D','A','B','C','10', 1);
    rev_round4_step($s--,'D','A','B','C','10', 2);
    rev_round4_step($s,  'A','B','C','D', '6', 0);
    rev_round4_step($s,  'A','B','C','D', '6', 1);
    rev_round4_step($s--,'A','B','C','D', '6', 2);
    rev_round4_step($s,  'B','C','D','A','21', 0);
    rev_round4_step($s,  'B','C','D','A','21', 1);
    rev_round4_step($s--,'B','C','D','A','21', 2);
    rev_round4_step($s,  'C','D','A','B','15', 0);
    rev_round4_step($s,  'C','D','A','B','15', 1);
    rev_round4_step($s--,'C','D','A','B','15', 2);
    rev_round4_step($s,  'D','A','B','C','10', 0);
    rev_round4_step($s,  'D','A','B','C','10', 1);
    rev_round4_step($s--,'D','A','B','C','10', 2);
    rev_round4_step($s,  'A','B','C','D', '6', 0);
    rev_round4_step($s,  'A','B','C','D', '6', 1);
    rev_round4_step($s--,'A','B','C','D', '6', 2);
    rev_round4_step($s,  'B','C','D','A','21', 0);
    rev_round4_step($s,  'B','C','D','A','21', 1);
    rev_round4_step($s--,'B','C','D','A','21', 2);
    rev_round4_step($s,  'C','D','A','B','15', 0);
    rev_round4_step($s,  'C','D','A','B','15', 1);
    rev_round4_step($s--,'C','D','A','B','15', 2);
    rev_round4_step($s,  'D','A','B','C','10', 0); # step 49
    rev_round4_step($s,  'D','A','B','C','10', 1);
    rev_round4_step($s--,'D','A','B','C','10', 2);
    # subsequent step (step 48) would need access to X[0] which is not known
    $code .= <<EOF;
  ; save target values D for the two interleaved MD5
  mov $targetd1, $d1
  mov $targetd2, $d2
  mov $targetd3, $d3

  ; init counter
  mov r4.x, l100.x

  whileloop
    ; break if counter is 0
    break_logicalz r4.x
    ; decrement loop counter
    iadd r4.x, r4.x, l100.z
    ; init A,B,C,D for first interleaved MD5
    mov r0, l101.xxxx
    mov r1, l101.yyyy
    mov r2, l101.zzzz
    mov r3, l101.wwww
    ; init A,B,C,D for second interleaved MD5
    mov r24, r0
    mov r25, r1
    mov r26, r2
    mov r27, r3
    ; init A,B,C,D for third interleaved MD5
    mov r31, r0
    mov r32, r1
    mov r33, r2
    mov r34, r3
EOF
    $s = 0;

    round1_step($s,  'A','B','C','D', '7', 0);
    round1_step($s,  'A','B','C','D', '7', 1);
    round1_step($s++,'A','B','C','D', '7', 2);
    round1_step($s,  'D','A','B','C','12', 0);
    round1_step($s,  'D','A','B','C','12', 1);
    round1_step($s++,'D','A','B','C','12', 2);
    round1_step($s,  'C','D','A','B','17', 0);
    round1_step($s,  'C','D','A','B','17', 1);
    round1_step($s++,'C','D','A','B','17', 2);
    round1_step($s,  'B','C','D','A','22', 0);
    round1_step($s,  'B','C','D','A','22', 1);
    round1_step($s++,'B','C','D','A','22', 2);
    round1_step($s,  'A','B','C','D', '7', 0);
    round1_step($s,  'A','B','C','D', '7', 1);
    round1_step($s++,'A','B','C','D', '7', 2);
    round1_step($s,  'D','A','B','C','12', 0);
    round1_step($s,  'D','A','B','C','12', 1);
    round1_step($s++,'D','A','B','C','12', 2);
    round1_step($s,  'C','D','A','B','17', 0);
    round1_step($s,  'C','D','A','B','17', 1);
    round1_step($s++,'C','D','A','B','17', 2);
    round1_step($s,  'B','C','D','A','22', 0);
    round1_step($s,  'B','C','D','A','22', 1);
    round1_step($s++,'B','C','D','A','22', 2);
    round1_step($s,  'A','B','C','D', '7', 0);
    round1_step($s,  'A','B','C','D', '7', 1);
    round1_step($s++,'A','B','C','D', '7', 2);
    round1_step($s,  'D','A','B','C','12', 0);
    round1_step($s,  'D','A','B','C','12', 1);
    round1_step($s++,'D','A','B','C','12', 2);
    round1_step($s,  'C','D','A','B','17', 0);
    round1_step($s,  'C','D','A','B','17', 1);
    round1_step($s++,'C','D','A','B','17', 2);
    round1_step($s,  'B','C','D','A','22', 0);
    round1_step($s,  'B','C','D','A','22', 1);
    round1_step($s++,'B','C','D','A','22', 2);
    round1_step($s,  'A','B','C','D', '7', 0);
    round1_step($s,  'A','B','C','D', '7', 1);
    round1_step($s++,'A','B','C','D', '7', 2);
    round1_step($s,  'D','A','B','C','12', 0);
    round1_step($s,  'D','A','B','C','12', 1);
    round1_step($s++,'D','A','B','C','12', 2);
    round1_step($s,  'C','D','A','B','17', 0);
    round1_step($s,  'C','D','A','B','17', 1);
    round1_step($s++,'C','D','A','B','17', 2);
    round1_step($s,  'B','C','D','A','22', 0);
    round1_step($s,  'B','C','D','A','22', 1);
    round1_step($s++,'B','C','D','A','22', 2);

    round2_step($s,  'A','B','C','D', '5', 0);
    round2_step($s,  'A','B','C','D', '5', 1);
    round2_step($s++,'A','B','C','D', '5', 2);
    round2_step($s,  'D','A','B','C', '9', 0);
    round2_step($s,  'D','A','B','C', '9', 1);
    round2_step($s++,'D','A','B','C', '9', 2);
    round2_step($s,  'C','D','A','B','14', 0);
    round2_step($s,  'C','D','A','B','14', 1);
    round2_step($s++,'C','D','A','B','14', 2);
    round2_step($s,  'B','C','D','A','20', 0);
    round2_step($s,  'B','C','D','A','20', 1);
    round2_step($s++,'B','C','D','A','20', 2);
    round2_step($s,  'A','B','C','D', '5', 0);
    round2_step($s,  'A','B','C','D', '5', 1);
    round2_step($s++,'A','B','C','D', '5', 2);
    round2_step($s,  'D','A','B','C', '9', 0);
    round2_step($s,  'D','A','B','C', '9', 1);
    round2_step($s++,'D','A','B','C', '9', 2);
    round2_step($s,  'C','D','A','B','14', 0);
    round2_step($s,  'C','D','A','B','14', 1);
    round2_step($s++,'C','D','A','B','14', 2);
    round2_step($s,  'B','C','D','A','20', 0);
    round2_step($s,  'B','C','D','A','20', 1);
    round2_step($s++,'B','C','D','A','20', 2);
    round2_step($s,  'A','B','C','D', '5', 0);
    round2_step($s,  'A','B','C','D', '5', 1);
    round2_step($s++,'A','B','C','D', '5', 2);
    round2_step($s,  'D','A','B','C', '9', 0);
    round2_step($s,  'D','A','B','C', '9', 1);
    round2_step($s++,'D','A','B','C', '9', 2);
    round2_step($s,  'C','D','A','B','14', 0);
    round2_step($s,  'C','D','A','B','14', 1);
    round2_step($s++,'C','D','A','B','14', 2);
    round2_step($s,  'B','C','D','A','20', 0);
    round2_step($s,  'B','C','D','A','20', 1);
    round2_step($s++,'B','C','D','A','20', 2);
    round2_step($s,  'A','B','C','D', '5', 0);
    round2_step($s,  'A','B','C','D', '5', 1);
    round2_step($s++,'A','B','C','D', '5', 2);
    round2_step($s,  'D','A','B','C', '9', 0);
    round2_step($s,  'D','A','B','C', '9', 1);
    round2_step($s++,'D','A','B','C', '9', 2);
    round2_step($s,  'C','D','A','B','14', 0);
    round2_step($s,  'C','D','A','B','14', 1);
    round2_step($s++,'C','D','A','B','14', 2);
    round2_step($s,  'B','C','D','A','20', 0);
    round2_step($s,  'B','C','D','A','20', 1);
    round2_step($s++,'B','C','D','A','20', 2);

    round3_step($s,  'A','B','C','D', '4', 0);
    round3_step($s,  'A','B','C','D', '4', 1);
    round3_step($s++,'A','B','C','D', '4', 2);
    round3_step($s,  'D','A','B','C','11', 0);
    round3_step($s,  'D','A','B','C','11', 1);
    round3_step($s++,'D','A','B','C','11', 2);
    round3_step($s,  'C','D','A','B','16', 0);
    round3_step($s,  'C','D','A','B','16', 1);
    round3_step($s++,'C','D','A','B','16', 2);
    round3_step($s,  'B','C','D','A','23', 0);
    round3_step($s,  'B','C','D','A','23', 1);
    round3_step($s++,'B','C','D','A','23', 2);
    round3_step($s,  'A','B','C','D', '4', 0);
    round3_step($s,  'A','B','C','D', '4', 1);
    round3_step($s++,'A','B','C','D', '4', 2);
    round3_step($s,  'D','A','B','C','11', 0);
    round3_step($s,  'D','A','B','C','11', 1);
    round3_step($s++,'D','A','B','C','11', 2);
    round3_step($s,  'C','D','A','B','16', 0);
    round3_step($s,  'C','D','A','B','16', 1);
    round3_step($s++,'C','D','A','B','16', 2);
    round3_step($s,  'B','C','D','A','23', 0);
    round3_step($s,  'B','C','D','A','23', 1);
    round3_step($s++,'B','C','D','A','23', 2);
    round3_step($s,  'A','B','C','D', '4', 0);
    round3_step($s,  'A','B','C','D', '4', 1);
    round3_step($s++,'A','B','C','D', '4', 2);
    round3_step($s,  'D','A','B','C','11', 0);
    round3_step($s,  'D','A','B','C','11', 1);
    round3_step($s++,'D','A','B','C','11', 2);
    round3_step($s,  'C','D','A','B','16', 0);
    round3_step($s,  'C','D','A','B','16', 1);
    round3_step($s++,'C','D','A','B','16', 2);
    round3_step($s,  'B','C','D','A','23', 0);
    round3_step($s,  'B','C','D','A','23', 1);
    round3_step($s++,'B','C','D','A','23', 2);
    round3_step($s,  'A','B','C','D', '4', 0);
    round3_step($s,  'A','B','C','D', '4', 1);
    round3_step($s++,'A','B','C','D', '4', 2);
    round3_step($s,  'D','A','B','C','11', 0);
    round3_step($s,  'D','A','B','C','11', 1);
    round3_step($s++,'D','A','B','C','11', 2);
    # can stop after step 45 (out of 0..63) since we have computed D.
    # the next steps in round 3 would have been:
    #round3_step($s,  'C','D','A','B','16');
    #round3_step($s++,'C','D','A','B','16', 1);
    #round3_step($s,  'B','C','D','A','23');
    #round3_step($s++,'B','C','D','A','23', 1);

    $code .= <<EOF;

    ; set bits in r5 if one of the interleaved MD5 found the target value D
    ieq r5, $d1, $targetd1
    ieq r6, $d2, $targetd2
    ior r5, r5, r6
    ieq r6, $d3, $targetd3
    ior r5, r5, r6

    iadd $firstword1.x, $firstword1.x, l100.w ; one
    iand r6.x, $firstword1.x, l104.x
    ieq r6.x, r6.x, l104.y
EOF
    if ($pos >= 16) {
	$code .= <<EOF;
    if_logicalnz r6.x
      iadd $firstword1.x, $firstword1.x, l104.z
      iand r6.x, $firstword1.x, l105.x
      ieq r6.x, r6.x, l105.y
EOF
	if ($pos >= 24) {
	    $code .= <<EOF;
      if_logicalnz r6.x
        iadd $firstword1.x, $firstword1.x, l105.z
        iand r6.x, $firstword1.x, l106.x
        ieq r6.x, r6.x, l106.y
EOF
	    if ($pos >= 32) {
		$code .= <<EOF;
        if_logicalnz r6.x
          iadd $firstword1.x, $firstword1.x, l106.z
          iand r6.x, $firstword1.x, l107.x
          ieq r6.x, r6.x, l107.y
EOF
	    }
	}
    }
    $code .= <<EOF;
    ; if true, we have reached the end of the work unit, set bits in r5
    ior r5, r5, r6.xxxx
EOF
    $code .= "        endif\n" if $pos >= 32;
    $code .= "      endif\n" if $pos >= 24;
    $code .= "    endif\n" if $pos >= 16;
    $code .= <<EOF;
    iand r6.x, $firstword1.x, l103.x
    iand $firstword1, $firstword1, l109.xxxx
    iand $firstword2, $firstword2, l109.xxxx
    iand $firstword3, $firstword3, l109.xxxx
    ior $firstword1, $firstword1, r6.xxxx
    ior $firstword2, $firstword2, r6.xxxx
    ior $firstword3, $firstword3, r6.xxxx

    ; if r5 has any bit set, break
    ior r5.xy, r5.xy, r5.zw
    ior r5.x, r5.x, r5.y
    break_logicalnz r5.x
  endloop

EOF

    my @components = qw/x y z w/;
    # build the names of the 'sizeof (thread_state_t) / 4' words of state that
    # are going to be written to g[]
    my @gwords = ();
    for (my $i = 0; $i < 3 * $elm_per_threads; $i++) {
	push @gwords, sprintf("g[r4.y+%d].%s", $i / 4, $components[$i % 4]);
    }

    for (my $interleave = 0; $interleave < $elm_per_threads / 4;
	$interleave++) {
	my $id = $interleave + 1;
	my $d = eval "\$d$id";
	my $targetd = eval "\$targetd$id";
	my $firstword = eval "\$firstword$id";
	my $secondword = eval "\$secondword$id";
	my $lit;
	$lit = 'l105' if $pos == 16;
	$lit = 'l106' if $pos == 24;
	$lit = 'l107' if $pos == 32;
	$code .= <<EOF;
  ; analyze results of MD5 interleave #$interleave, and
  ; store the 4 elements' states in global buffer
  ieq r5, $d, $targetd        ; found D -> set bits in r5
  iand r6, $firstword, $lit.xxxx
  ieq r6, r6, $lit.yyyy ; completed work unit -> set bits in r6

EOF
	foreach my $elm (0 .. 3) {
	    my $compo = $components[$elm];
	    my $status = shift @gwords;
	    my $wuid0 = shift @gwords;
	    my $wuid1 = shift @gwords;
	    $code .= <<EOF;
  ; elm $elm ($compo)
  ixor r0.x, $status, l103.w
  if_logicalnz r0.x
    mov $status, l103.y
    if_logicalnz r5.$compo
      mov $status, l100.w
    else
      if_logicalnz r6.$compo
        mov $status, l103.z
      endif
    endif
    mov $wuid0, $firstword.$compo
    mov $wuid1, $secondword.$compo
  endif

EOF
	}
    }

    $code .= ";_the_end_\nend\n";
# $code contains a compute shader program:
#   il_cs
#     ...
#   end
# now it is transformed to a C macro string:
#   #define KERNEL \
#   "il_cs\
#     ...\
#   end\
#   "
    chomp $code;
    $code =~ s/"/\\"/g;
    $code =~ s/$/\\n\\/gm;
    my $macro_name = $fname;
    $macro_name =~ s/\.h$//;
    $macro_name =~ s/-/_/g;
    $macro_name = uc($macro_name);
    $code = "#define $macro_name \\\n\"$code\n\"\n";
    my $fh;
    open($fh, ">", $fname) or die "can't open $fname: $!";
    print { $fh } $code;
    close($fh) or die "can't close $fname: $!";
}

generate_kernel("kernel-md5-pos16-nobfi.h", 16, 0);
generate_kernel("kernel-md5-pos16-bfi.h", 16, 1);
generate_kernel("kernel-md5-pos24-nobfi.h", 24, 0);
generate_kernel("kernel-md5-pos24-bfi.h", 24, 1);
generate_kernel("kernel-md5-pos32-nobfi.h", 32, 0);
generate_kernel("kernel-md5-pos32-bfi.h", 32, 1);
# eof
